/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <assert.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "basics.h"
#include "dom_x.h"
#include "tools.h"
#include "make_x.h"
#include "petrinet_x.h"

#include "cxp_dtd.h"


/****************************************************************************

  type definition: 
*/
typedef enum {
  state,
  transition
} knot_type_t;


/* ***************************************************************************

  type definition: knot_t
 */

/*! knot structure
 */
typedef struct {
  knot_type_t type;             /*!< type of element */
  xmlChar *id;			/*!< id of knot */
  BOOL_T out;			/*!< output marker flag */
  index_t i;                    /*!< index in array, redundant but performant */
  index_t d;			/*!< minimum edge distance between p->start and this knot */
  xmlNodePtr node;		/*!< ptr to the according DOM node */
} knot_t;


/* ***************************************************************************

   preprocessor macros for access to knot_t
 */

#define statep(N) (N != NULL && N->type == state)

#define transitionp(N) (N != NULL && N->type == transition)

/* warning: not nice -> better edgep() */

#define edge_st(P,s,t) (P->e[s][t] & 1)

#define edge_ts(P,t,s) (P->e[s][t] & 2)

#define knotid(N) N->id

#define knotname(N) N->name

#define state_index(N) (statep(N) ? N->i : ERROR_INDEX)

#define transition_index(N) (transitionp(N) ? N->i : ERROR_INDEX)

/****************************************************************************

  type definition: edge_t
*/
typedef struct {
  knot_t *from;			/*!< ptr to origin knot of edge */
  knot_t *to;			/*!< ptr to target knot of edge */
  BOOL_T out;			/*!< output marker flag */
  xmlNodePtr node;		/*!< ptr to the according DOM node */
} edge_t;


/****************************************************************************

  type definition: petrinet_t
*/

typedef struct {
  xmlDocPtr xml;			/*!< ptr to XML DOM of libxml2 */
  xmlNode *root_element;	/*!< ptr to the root element of DOM */
  xmlChar *id;			/*!< id of petrinet */
#if 0
  xmlNode *cfg_root_element;
  xmlChar *version;
  xmlChar *name;
  xmlChar *description;
  xmlChar *copyright;
  xmlChar *remark;
  xmlChar *contact;
#endif
  knot_t *knot;			/*!< ptr to array of knots */

  index_t n_s;
  knot_t *s;
  index_t n_t;
  knot_t *t;
/*   index_t n_k;			/\* number of knots *\/ */
/*   knot_t **k; */
  index_t n_e;			/* number of edges */
  edge_t *edge;			/* array of edges */
  index_t **e;                  /* 2D matrix of edges */
  knot_t **path;                        /* pointer for path buffer */
  index_t n_paths; /* number of found path between 'start' and 'target' */
  index_t l_path;               /* last knot index in path */
  index_t l_min;		/* minimal length of valid path */
  index_t l_max;		/* maxmal length of valid path */
  knot_t *start;
  knot_t *target;
/*   char doc_prefix[BUFFER_LENGTH]; */
/*   char *html_prefix; */
/*   char *executable; */
/*   char *filename;		/\* name of parsed cfg xml file -> CGI *\/ */
/*   char *filename_absolut;		/\* name of parsed cfg xml file -> CGI *\/ */
} petrinet_t;


/****************************************************************************

   preprocessor macros for access to petrinet_t
*/


/****************************************************************************

   Declaration of public constants
*/

extern const xmlChar *tag_state;

extern const xmlChar *tag_transition;

extern const xmlChar *tag_edge;


/****************************************************************************

   preprocessor macros for access to petrinet_t
*/

#define petrinet_set_start(P,ID) set_start_target(P,FALSE,ID)

#define petrinet_set_target(P,ID) set_start_target(P,TRUE,ID)

#define mode_cgi(P) (P->mode_run == cgi)


/****************************************************************************

   Definition of private constants
*/

const xmlChar *tag_pnet = BAD_CAST  "PETRINET";

const xmlChar *tag_subnet = BAD_CAST  "SUBNET";

const xmlChar *tag_state = BAD_CAST  "STELLE";

const xmlChar *tag_transition = BAD_CAST  "TRANSITION";

const xmlChar *tag_edge = BAD_CAST  "RELATION";

/****************************************************************************

   Declaration of private functions (prototypes)
*/

static petrinet_t *
petrinet_new(xmlDocPtr xml);

static BOOL_T 
set_start_target(petrinet_t *p, BOOL_T flag_end, xmlChar *node_id);

static BOOL_T 
path_forward(petrinet_t *p);

static BOOL_T 
branch_backward(petrinet_t *p);

static int 
path_next(void);

static index_t 
state_index4id(petrinet_t *p, xmlChar *id);

static index_t 
transition_index4id(petrinet_t *p, xmlChar *id);

static index_t
edge_index4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to);

static edge_t *
edge_ptr4ptr(petrinet_t *p, knot_t *ptr_from, knot_t *ptr_to);

static void
petrinet_free(petrinet_t *p);

static index_t
get_knot_distance(petrinet_t *p, knot_t *a, knot_t *b);
/* get_knot_distance(petrinet_t *p, xmlChar *start_id, xmlChar *target_id); */

static BOOL_T
petrinet_center_valid(petrinet_t *p, xmlChar *s_id);

/* static BOOL_T */
/* petrinet_pathnet_out_flag(petrinet_t *p, BOOL_T f); */

static BOOL_T
set_valid_out(petrinet_t *p);

static void
petrinet_path_out_flag(petrinet_t *p, BOOL_T f);

static void
petrinet_all_out_flag(petrinet_t *p, BOOL_T f);

static void
petrinet_neighbors_out_flag(petrinet_t *p, knot_t *k, BOOL_T f);

static void
petrinet_cascading_out_flag(petrinet_t *p, knot_t *k, BOOL_T f, index_t depth);

static knot_t *
state_ptr4id(petrinet_t *p, xmlChar *id);

static knot_t *
transition_ptr4id(petrinet_t *p, xmlChar *id);

static edge_t *
edge_ptr4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to);

static BOOL_T
set_path_length(petrinet_t *p, char *str_min, char *str_max);

static xmlNodePtr
xml_path(petrinet_t *p, xmlNodePtr parent, xmlNsPtr ns_xsl);

static xmlDocPtr
xml_pathnet(petrinet_t *p);

static xmlDocPtr
xml_pathtable(petrinet_t *p);

static BOOL_T
petrinet_states_new(petrinet_t *p);

static BOOL_T
petrinet_transitions_new(petrinet_t *p);

static BOOL_T
petrinet_edges_new(petrinet_t *p);

static edge_t *
petrinet_edges_init(petrinet_t *p, edge_t *current_edge, xmlNode * a_node);

static knot_t *
petrinet_knots_init(petrinet_t *p, knot_t *current_knot, xmlNode *a_node, const xmlChar *element_name);

static BOOL_T 
member_transition(petrinet_t *p, knot_t *t) ;

static BOOL_T 
member_state(petrinet_t *p, knot_t *s) ;

static BOOL_T
new_knots(petrinet_t *p);

static knot_t *
init_knots(const petrinet_t *p, knot_t *current_knot, xmlNode *a_node);

/****************************************************************************

   Definition of all functions
*/


/****************************************************************************

   Definition of functions
*/


/*! process the required  files
 */
xmlDocPtr
process_path(xmlNodePtr node_path)
{
  xmlNodePtr node_xml;
  xmlDocPtr result = NULL;
  xmlDocPtr doc_pnet;
  xmlChar *attr_start;
  xmlChar *attr_target;
  petrinet_t *p;

  printf_log(1,"Processing PATH");

  attr_start = attribute_str(node_path,BAD_CAST "start");
  attr_target = attribute_str(node_path,BAD_CAST "target");

  if (((attr_start && xmlStrlen(attr_start)) || (attr_target && xmlStrlen(attr_target)))
      && (node_xml = child_element(node_path,NAME_XML))) {
    /*  */
    doc_pnet = process_xml(node_xml); /*!\bug apply XSL?? */
    if (doc_pnet) {
      p = petrinet_new(doc_pnet);

      set_path_length(p,
		      (char *) attribute_str(node_path, BAD_CAST "lmin"),
		      (char *) attribute_str(node_path, BAD_CAST "lmax"));

      printf_exit( attr_start && !petrinet_set_start(p, attr_start),
		   41, "Cant set start STELLE" );

      /* target knot */
      printf_exit( attr_target && !petrinet_set_target(p, attr_target),
		   42, "Cant set target STELLE" );

      /*!\todo insert meta element */

      if (xmlStrcasecmp(node_path->name,BAD_CAST "pathtable")==0) {
	result = xml_pathtable(p);
      }
      else {
	result = xml_pathnet(p);
      }

      petrinet_free(p);
      xmlFreeDoc(doc_pnet);
    }
  }
  else {
    printf_log(1,"No valid attributes in '%s'",node_path->name);
  }

  return result;
}


/*!

\param xml DOM of PETRINET

\return the pointer of the created petrinet
*/
petrinet_t *
petrinet_new(xmlDocPtr xml)
{
  petrinet_t *result;

  result = (petrinet_t *) xmlMemMalloc(sizeof(petrinet_t));
  printf_exit((result == NULL), 99, "No more memory");

  result->xml = xml;

  /* Get the root element node */
  result->root_element = xmlDocGetRootElement(result->xml);

  /* set id of petrinet */
  result->id = attribute_str(result->root_element, BAD_CAST  "ID");
  if (result->id==NULL) {
    /* no valid id set */
    result->id = xmlStrdup(BAD_CAST  "anonymous");
  }

  /* 
  */
  new_knots(result);

  petrinet_states_new(result);

  petrinet_transitions_new(result);

  petrinet_edges_new(result);

  result->path = (knot_t **) xmlMemMalloc((result->n_s + result->n_t + 1) * sizeof(knot_t*));
  printf_exit((result->path == NULL), 99, "No more memory");

  result->l_path = ERROR_INDEX;

  result->l_min = ERROR_INDEX;
  result->l_max = ERROR_INDEX;

  result->start = NULL;

  result->target = NULL;

  return result;
}
/* End of petrinet_new() */


/*!

\param p pointer to the petrinet

\return success flag
*/
static BOOL_T
petrinet_states_new(petrinet_t *p)
{
  knot_t *tail;
  int i;

  assert(p->root_element);

  /* states */
  p->n_s = number_of(p->root_element, tag_state, 0);

  printf_log(1,"Number of '%s' = %i", tag_state, p->n_s);

  p->s = (knot_t *) xmlMemMalloc((p->n_s+1) * sizeof(knot_t));
  printf_exit((p->s == NULL), 99, "No more memory");

  tail = petrinet_knots_init(p, p->s, p->root_element, tag_state);

  tail->id = NULL;              /* mark the end of state array */

  /* set the index number */
  for (i=0; p->s[i].id; i++) {
    assert (i < p->n_s);
    p->s[i].i = i;
  }

  return TRUE;
}
/* End of petrinet_states_new() */


/*! 

\param p pointer to the petrinet

\return success flag
*/
static BOOL_T
petrinet_transitions_new(petrinet_t *p)
{
  knot_t *tail;
  int i;

  assert(p->root_element);

  /* transitions */
  p->n_t = number_of(p->root_element, tag_transition, 0);

  printf_log(1,"Number of '%s' = %i", tag_transition, p->n_t);

  p->t = (knot_t *) xmlMemMalloc((p->n_t+1) * sizeof(knot_t));
  printf_exit((p->t == NULL), 99, "No more memory");

  tail = petrinet_knots_init(p, p->t, p->root_element, tag_transition);

  tail->id = NULL;              /* mark the end of transition array */

  /* set the index number */
  for (i=0; p->t[i].id; i++) {
    assert (i < p->n_t);
    p->t[i].i = i;
  }

  return TRUE;
}
/* End of petrinet_transitions_new() */


/*!

\param p pointer to the petrinet

\return success flag
*/
static BOOL_T
petrinet_edges_new(petrinet_t *p)
{
  index_t i;
  index_t j;
  /* index_t **ptr;   */         /* pointer to new allocated memory */
  index_t *ptr_b;
  edge_t *tail;

  assert(p->root_element);
  assert(p->n_s > 0 && p->n_t > 0);

  /* block for matrix
   */
  ptr_b = (index_t *)xmlMemMalloc(p->n_s * p->n_t * sizeof(index_t));
  printf_exit((ptr_b == NULL), 99, "No more memory");

  /* array of line pointers 
   */
  p->e = (index_t **)xmlMemMalloc(p->n_s * sizeof(index_t *));
  printf_exit((p->e == NULL), 99, "No more memory");
  for (j=0; j<p->n_s; j++) {
    p->e[j] = ptr_b + j * p->n_t;
    /* set all possible edges to 'none' */
    for (i=0; i < p->n_t; i++) {
      p->e[j][i] = 0;
      /*!\todo optimize with memset(ptr_b) ?? */
    }
  }

  /* edges */
  p->n_e = number_of(p->root_element, tag_edge, 0);
  printf_log(1,"Number of '%s' = %i", tag_edge, p->n_e);
  if (p->n_e > 0) {
    p->edge = (edge_t *) xmlMemMalloc( (p->n_e+1) * sizeof(edge_t));
    printf_exit((p->edge == NULL), 99, "No more memory");

    tail = petrinet_edges_init(p, p->edge, p->root_element);
    tail->from = NULL;
    p->n_e = (tail - p->edge);
  }

  printf_log(1,"%i valid Edges", p->n_e);
  if (p->n_e > 0) {
    return TRUE;
  }
  else {
    return FALSE;
  }
}
/* End of petrinet_edges_new() */


/*! initialize all found edges

\param p pointer to the petrinet
\param current_edge
\param a_node current node for recursive call

\return 
*/
static edge_t *
petrinet_edges_init(petrinet_t *p, edge_t *current_edge, xmlNode * a_node)
{
  xmlNode *current_node;

  for (current_node = a_node; current_node; current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
        && !xmlStrcasecmp(current_node->name, BAD_CAST  "RELATION")) {
      /* current_node is a relation node */
      index_t a;
      index_t b;
      xmlChar *a_str;
      xmlChar *b_str;

      current_edge->node = current_node;
      current_edge->out = TRUE;
      a_str = attribute_str(current_node, BAD_CAST  "FROM");
      b_str = attribute_str(current_node, BAD_CAST  "TO");

      if (a_str && b_str) {
	a = state_index4id(p, a_str);
	if (a == ERROR_INDEX) {
	  /* it was a wrong assumtion,
	     OR there is no attribute 'FROM'
	     OR there is no valid stelle with id 'a_str'
	  */
	  a = transition_index4id(p, a_str);
	  if (a == ERROR_INDEX) {
	    /* ignore invalid relations */
	    /* 	    printf_log(1,"Edge invalid FROM '%s' -> '%s'", a_str, b_str); */
	  }
	  else {
	    b = state_index4id(p, b_str);
	    if (b != ERROR_INDEX) {
	      /* b is a valid stelle, a is a valid transition */
	      p->e[b][a] |= 2;

	      current_edge->from  = &(p->t[a]);
	      current_edge->to    = &(p->s[b]);
	      current_edge++;
	    }
	  }
	}
	else {
	  /* right assumtion */
	  b = transition_index4id(p, b_str);

	  if (b != ERROR_INDEX) {
	    /* a is a valid stelle, b is a valid transition */
	    p->e[a][b] |= 1;

	    current_edge->from  = &(p->s[a]);
	    current_edge->to    = &(p->t[b]);
	    current_edge++;
	  }
	  else {
	    /* ignore invalid relations */
	    /* 	      printf_log(1,"Edge invalid TO '%s' -> '%s'", a_str, b_str); */
	  }
	}
      }
    }
    current_edge = petrinet_edges_init(p,current_edge,current_node->children);
  }

  return current_edge;
}


/*! allocate and call init of knots array

\param p pointer to the petrinet

\return success flag
*/
static BOOL_T
new_knots(petrinet_t *p)
{
  knot_t *tail;
  int i;

  assert(p->root_element);

  i = number_of(p->root_element, tag_state, 0)
    + number_of(p->root_element, tag_transition, 0);

  printf_log(1,"Number of '%s' + '%s' = %i", tag_state, tag_transition, i);

  p->knot = (knot_t *) xmlMemMalloc((i+1) * sizeof(knot_t));
  printf_exit((p->knot == NULL), 99, "No more memory");

  tail = init_knots(p, p->knot, p->root_element);

/*   assert(tail - p->knot == i); */

  tail->id = NULL;              /* mark the end of knot array */

  /* set the index number */
  for (i=0; p->knot[i].id; i++) {
    p->knot[i].i = i;
  }

  return TRUE;
}
/* End of new_knots() */


/*! initialize all found knots

\param p pointer to the petrinet
\param current_knot
\param a_node current node for recursive call

\return 
*/
static knot_t *
init_knots(const petrinet_t *p, knot_t *current_knot, xmlNode *a_node)
{
  xmlNode *current_node;

  for (current_node = a_node; current_node; current_node = current_node->next) {
    if (current_node->type == XML_ELEMENT_NODE) {
      if (!xmlStrcasecmp(current_node->name, tag_state)) {
	current_knot->type = state;
      }
      else if (!xmlStrcasecmp(current_node->name, tag_transition)) {
	current_knot->type = transition;
      }
      else {
	continue;
      }

      /* set id of knot */
      current_knot->id = attribute_str(current_node, BAD_CAST  "ID");

      current_knot++;
    }

    current_knot = init_knots(p, current_knot, current_node->children);
  }

  return current_knot;
}
/* End of init_knots() */


/*! initialize all knots

\param p pointer to the petrinet
\param current_knot pointer for recursion
\param a_node
\param element_name tag name

\return 
*/
static knot_t *
petrinet_knots_init(petrinet_t *p, knot_t *current_knot, xmlNode *a_node, const xmlChar *element_name)
{
  xmlNode *current_node;

  for (current_node = a_node; current_node; current_node = current_node->next) {

    if (current_node->type == XML_ELEMENT_NODE
	&& !xmlStrcasecmp(current_node->name, element_name)) {
      /* current_node is a node w/ element_name */

      if (!xmlStrcasecmp(current_node->name, tag_state)) {
	current_knot->type = state;
      }
      else if (!xmlStrcasecmp(current_node->name, tag_transition)) {
	current_knot->type = transition;
      }
      else {
	continue;
      }

      current_knot->node = current_node;

      /* set id of knot */
      current_knot->id = attribute_str(current_node, BAD_CAST  "ID");

      current_knot++;
    }

    current_knot =
      petrinet_knots_init(p, current_knot, current_node->children, element_name);
  }

  return current_knot;
}
/* End of petrinet_knots_init() */

/*! release the allocated mem of a petrinet_t
*/
void
petrinet_free(petrinet_t *p)
{

  xmlMemFree(p->s);

  xmlMemFree(p->t); 

/*   xmlMemFree(p->sn); */

  xmlMemFree(p->edge);

#if 0
  mem_free_2d((void **)p->e);
#else
  xmlMemFree(*p->e);               /* release the block */
  xmlMemFree(p->e);                /* release the pointer array */
#endif

  xmlMemFree(p->knot);
/*   xmlMemFree(p->k); */
  xmlMemFree(p->path);

  xmlMemFree(p);

  return;
}
/* End of petrinet_free() */


/*! searches the index number of 
 */
index_t
edge_index4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to)
{
  index_t i;

  if (p && id_from && id_to) { 
    for (i=0; i < p->n_e; i++) {
      if ((!xmlStrcasecmp(p->edge[i].from->id, id_from) && !xmlStrcasecmp(p->edge[i].to->id, id_to))) {
	return i;
      }
    }
  } 

  return ERROR_INDEX;
}
/* End of edge_index4id() */


/*! searches the index number of 
 */
edge_t *
edge_ptr4ptr(petrinet_t *p, knot_t *ptr_from, knot_t *ptr_to)
{
  index_t i;

  if (p && ptr_from && ptr_to) { 
    for (i=0; i < p->n_e; i++) {
      if (p->edge[i].from == ptr_from
	  && p->edge[i].to   == ptr_to) {
	return &(p->edge[i]);
      }
    }
  }

  return NULL;
}
/* End of edge_index4id() */


index_t
state_index4id(petrinet_t *p, xmlChar *id)
{
  /* state_index4id() - searches the index number of 'id' in 'p->s[]'
  */

  index_t i;

  if (p && id) { 
    for (i=0; i < p->n_s; i++) {
      if (!xmlStrcasecmp(p->s[i].id, id)) {
	return i;
      }
    }
  } 

  return ERROR_INDEX;
}
/* End of state_index4id() */


index_t
transition_index4id(petrinet_t *p, xmlChar *id)
{
  /* transition_index4id() - searches the index number of 'id' in 'p->t[]'
  */

  index_t i;

  if (p && id) {
    for (i=0; i < p->n_t; i++) {
      if (!xmlStrcasecmp(p->t[i].id, id)) {
	return i;
      }
    }
  }

  return ERROR_INDEX;
}
/* End of transition_index4id() */


knot_t *
state_ptr4id(petrinet_t *p, xmlChar *id) 
{
  if (p && id) {
    index_t i = state_index4id(p,id);
    if (i > -1 && i < p->n_s) {
      return &(p->s[i]);
    }
  }

  return NULL;
}


knot_t *
transition_ptr4id(petrinet_t *p, xmlChar *id) 
{
  if (p && id) {
    index_t i = transition_index4id(p,id);
    if (i > -1 && i < p->n_t) {
      return &(p->t[i]);
    }
  }

  return NULL;
}


edge_t *
edge_ptr4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to)
{
  if (p && id_from && id_to) {
    index_t i = edge_index4id(p,id_from,id_to);
    if (i > -1 && i < p->n_e) {
      return &(p->edge[i]);
    }
  }

  return NULL;
}


BOOL_T
edgep(petrinet_t *p, knot_t *a, knot_t *b)
{
  /* edgep() - 

  */

  assert(p);
  assert(a);
  assert(b);

  if (statep(a)) {
    if (transitionp(b)) {
      return (p->e[a->i][b->i] & 1);
    }
  }
  else if (transitionp(a)) {
    if (statep(b)) {
      return (p->e[b->i][a->i] & 2);
    }
  }

  return FALSE;
}
/* End of edgep() */


static BOOL_T
member_transition(petrinet_t *p, knot_t *t) 
{

  index_t i;

  assert( p );

  if (transitionp(t)) {
    for (i=0; i < p->l_path; i++) {
      if ( transitionp(p->path[i]) && p->path[i] == t ) {
        return TRUE;
      }
    }
  }

  return FALSE;
}


static BOOL_T
member_state(petrinet_t *p, knot_t *s) 
{

  index_t i;

  assert( p );

  if (statep(s)) {
    for (i=0; i < p->l_path; i++) {
      if ( statep(p->path[i]) && p->path[i] == s ) {
        return TRUE;
      }
    }
  }

  return FALSE;
}


/*! sets the 'p->start' or 'p->target'

a transition as start makes no sense !
*/
BOOL_T
set_start_target(petrinet_t *p, BOOL_T flag_end, xmlChar *node_id)
{
  index_t s;

  assert(p);

  if (node_id && xmlStrlen(node_id)) {
    s = state_index4id(p,node_id);
    if (s == ERROR_INDEX) {
      printf_log(1,"'State %s' not valid", node_id);
      return FALSE;
    }

    if (flag_end) {
      printf_log(1,"Set end to '%s' -> %i", node_id, s);
      p->target = &(p->s[s]);
    }
    else {
      printf_log(1,"Set start to '%s' -> %i", node_id, s);
      p->start = &(p->s[s]);    
    }
    p->path[0] = p->start;
  }
  else {
    /* value of node_id not defined */
    if (flag_end) {
      printf_log(1,"Set end to 'NULL'");
      p->target = NULL;
    }
    else {
      printf_log(1,"Set start to 'NULL'");
      p->start = NULL;    
    }
/*     p->path[0] = ; */
  }

  p->l_path  = 0;
  p->n_paths = 0;

  return TRUE;
}
/* End of set_start_target() */


/*! sets the 

*/
BOOL_T
set_path_length(petrinet_t *p, char *str_min, char *str_max)
{
  char *e;

  assert(p);

  if (str_min) {
    p->l_min = (index_t) strtol(str_min, &e, 10);
    if (!e || p->l_min < 1) {
      /* conversion error -> back to init value */
      p->l_min = ERROR_INDEX;
    }
    printf_log(1,"l_min = '%i'", p->l_min);
  }

  if (str_max) {
    p->l_max = (index_t) strtol(str_max, &e, 10);
    if (!e || p->l_max < 1) {
      /* conversion error -> back to init value */
      p->l_max = ERROR_INDEX;
    }
    printf_log(1,"l_max = '%i'", p->l_max);
  }

  return TRUE;
}
/* End of set_path_length() */


/*! search for an new path 

p->start and p->target must be states!! see set_start_target()
*/
BOOL_T
path_forward(petrinet_t *p)
{

  index_t i;
  knot_t *l_knot;

  /*! \bug bidrectional relation is also a path */

  /*   assert(statep(p->start) || transitionp(p->start)); */
  /*   assert(statep(p->target) || transitionp(p->target)); */
  assert(statep(p->start));
  assert(statep(p->target));
  assert(p->path[0] == p->start);

  l_knot = p->path[p->l_path]; /* ptr to last knot */

  if (p->l_max != ERROR_INDEX && p->l_path >= p->l_max) {
    /* path is too long */
    return FALSE;    /* leave the recursion */
  }
  else if (l_knot == p->target && p->l_path > 0) {
    /* target is reached, path is yet complete,
       probably branch_backward() found a short way */

    if (p->l_min != ERROR_INDEX && p->l_path < p->l_min) {
      /* path is too short */
      return FALSE;    /* leave the recursion */
    }

    p->n_paths++;   /* count this path and */
    return TRUE;    /* leave the recursion */
  }
  else if (statep(l_knot)) {
    /* last element is a state -> search for a transition knot */

    index_t l_state = state_index(l_knot);

    for (i=0; i < p->n_t; i++) { /* all transitions */
      if (edge_st(p,l_state,i)
          && (! member_transition(p, &(p->t[i])))) {
        /* found a valid transition,
	   with an edge to l_state and
	   not yet included in path */
        p->l_path++;
        p->path[p->l_path] = &(p->t[i]); /* enclose the pointer into path */
        return path_forward(p);	/* search recursive next knot for path[] */
      }
    }
  }
  else if (transitionp(l_knot)) {
    /* last element is a transition -> search for a state knot */

    index_t l_trans = transition_index(l_knot);

    for (i=0; i < p->n_s; i++) { /* all states */
      knot_t *s_ptr = &(p->s[i]);
      if (edge_ts(p,l_trans,i)) {
	/* found a valid state with an edge to l_trans */
	if (member_state(p, s_ptr)) {
#if 1
	  if (s_ptr == p->start && s_ptr == p->target) {
	    /* special case: its a circle! */
	    p->l_path++;
	    p->path[p->l_path] = s_ptr;
	    p->n_paths++;   /* count this path and */
	    return TRUE;    /* leave the recursion */
	  }
	  /* else: state included yet */
#endif
	}
	else {
	  /* not yet included in path */
	  p->l_path++;
	  p->path[p->l_path] = s_ptr;
	  if (p->path[p->l_path] == p->target) {
	    /* target is reached, path is completed */
	    p->n_paths++;   /* count this path and */
	    return TRUE;    /* leave the recusion */
	  }
	  return path_forward(p);	/* search recursive next knot for path[] */
	}
      }
    }
  }
  else {
    assert(FALSE);
  }

  return FALSE;			/* nothing more found */
}
/* End of path_forward() */


/*! search for the last possible branch 
 */
BOOL_T
branch_backward(petrinet_t *p)
{

  index_t i;
  knot_t *ol_knot;
  knot_t *l_knot;

  assert(statep(p->start));
  assert(statep(p->target));
  assert(p->path[0] == p->start);

  if (p->l_path < 1) {
    /* can't change the start, nothing more found */
    return FALSE;
  }

  ol_knot = p->path[p->l_path]; /* ptr to old last knot */

  p->l_path--;          /* one step backward */

  l_knot = p->path[p->l_path]; /* ptr to last knot */

  if (statep(l_knot)) {
    /* last element is a state -> search for a transition knot */

    index_t l_state = state_index(l_knot);
    index_t l_trans = transition_index(ol_knot); /* index of removed step */

    for (i=l_trans+1; i < p->n_t; i++) { /* next transitions after removed one */
      if (edge_st(p,l_state,i)
	  && ! member_transition(p, &(p->t[i]))) {
        /* found a valid transition,
	   with an edge to l_state and
	   not yet included in path */
        p->l_path++;
        p->path[p->l_path] = &(p->t[i]);
        return TRUE;		/* leave backward recursion */
      }
    }
    return branch_backward(p);
  }
  else {
    /* last element is a transition */

    index_t l_trans = transition_index(l_knot);
    index_t l_state = state_index(ol_knot);

    for (i=l_state+1; i < p->n_s; i++) {
      if (edge_ts(p,l_trans,i)
	  && ! member_state(p, &(p->s[i]))) {
        /* found a state */
        p->l_path++;
        p->path[p->l_path] = &(p->s[i]);
        return TRUE;
      }
    }
    return branch_backward(p);
  }

  assert(FALSE);
}
/* End of branch_backward() */


/*! computes the distance between two knots

\param p petrinet record
\param a pointer to start knot
\param b pointer to target knot

\return the lowest number of edges between a and b

 0 means no path between knots, or start_id is same as target_id
*/
index_t
get_knot_distance(petrinet_t *p, knot_t *a, knot_t *b)
{
  index_t result = 0;

  if (a != b) {
    p->start = a;
    p->target = b;

    /* find the minimum distance value */
    do {
      if (path_forward(p)) {
	if (result == 0 || p->l_path < result) {
	  result = p->l_path;
	}
      }
    } while (branch_backward(p));
  }

  return result;
}
/* End of get_knot_distance() */


void
petrinet_path_out_flag(petrinet_t *p, BOOL_T f)
{
  /* 
     set all knots in p->path on f
  */
  index_t i;
  edge_t *e;

  for (i=0; i <= p->l_path; i++) {
    p->path[i]->out = f;	/* set knot */
    if (i && (e = edge_ptr4ptr(p,p->path[i-1],p->path[i]))) {
      e->out = f;		/* set edge */
    }
  }
}
/* End of petrinet_path_out_flag() */


#if 0
/*! set all knots in paths between state  on f
*/
BOOL_T
petrinet_pathnet_out_flag(petrinet_t *p, BOOL_T f)
{

  if (p->start && p->target) {

    petrinet_all_out_flag(p,!f);

    do {
      if (path_forward(p)) {
	petrinet_path_out_flag(p,f);
      }
    } while (branch_backward(p));

    printf_log(1,"%i path between '%s' and '%s' found",
	    p->n_paths, p->start->id, p->target->id);
  }

  return (p->n_paths > 0);
}
/* End of petrinet_pathnet_out_flag() */
#endif


void
petrinet_all_out_flag(petrinet_t *p, BOOL_T f)
{

  knot_t *k;
  index_t i;

  assert(p);

  for (k=p->s; k->id; k++) {
    k->out = f;
  }

  for (k=p->t; k->id; k++) {
    k->out = f;
  }

  for (i=0; i < p->n_e; i++) {
    p->edge[i].out = f;
  }
}
/* End of petrinet_all_out_flag() */


/*! sets all VALID in DOM according to out flags in petrinet
*/
BOOL_T
set_valid_out(petrinet_t *p)
{
  knot_t *k;
  index_t i;

  assert(p);

  for (k=p->s; k->id; k++) {
    if (!k->out) {
      xmlSetProp(k->node, BAD_CAST "VALID", BAD_CAST "no");
    }
  }

  for (k=p->t; k->id; k++) {
    if (!k->out) {
      xmlSetProp(k->node, BAD_CAST "VALID", BAD_CAST "no");
    }
  }

  for (i=0; i < p->n_e; i++) {
    if (!p->edge[i].out) {
      xmlSetProp(p->edge[i].node, BAD_CAST "VALID", BAD_CAST "no");
    }
  }

  return TRUE;
}
/* end of set_valid_out() */


/*! compiles a DOM of pathnet

  \param p pointer to used petrinet

  \result complete DOM of pathnet

  \todo multithreading path searching (reentrant versions of path_forward() and branch_backward())

  \todo weight nodes according to the count of path members
*/
xmlDocPtr
xml_pathnet(petrinet_t *p)
{
  /* 
   */
  /*   int i = 0; */
  xmlNodePtr n_tmp;
  /*   xmlNodePtr pathnet; */
  xmlNodePtr stylesheet;
  /*   xmlNodePtr template_root; */
  xmlNodePtr template_xsl;
  xmlNodePtr template_choose;
  /*   xmlNodePtr attribute_count; */
  xmlNodePtr n_knot;
  xmlDocPtr result = NULL;
  xmlChar buffer[BUFFER_LENGTH];

  xmlNsPtr ns_xsl;

  knot_t *k;
  index_t i;

  petrinet_all_out_flag(p,FALSE);
  if (p->start) {
    if (p->target) {
      /* start and target are defined */
      do {
	if (path_forward(p)) {
	  petrinet_path_out_flag(p,TRUE);
	}
      } while (branch_backward(p));
      printf_log(1,"%i path between '%s' and '%s' found",
		   p->n_paths, p->start->id, p->target->id);
    }
    else {
      /* open target */
      for (k=p->s; k->id; k++) {
	if (k==p->start) {
	  /* no circles are allowed here */
	  printf_log(2,"Skip '%s'", k->id);
	}
	else {
	  /* target knot */
	  printf_exit( !petrinet_set_target(p, k->id), 42, "Cant set target STELLE" );
	  do {
	    if (path_forward(p)) {
	      petrinet_path_out_flag(p,TRUE);
	    }
	  } while (branch_backward(p));
	  printf_log(1,"%i path between '%s' and '%s' found",
		       p->n_paths, p->start->id, p->target->id);
	}
      }
    }
  }
  else {
    if (p->target) {
      /* open start */
      for (k=p->s; k->id; k++) {
	if (k==p->target) {
	  /* no circles are allowed here */
	  printf_log(2,"Skip '%s'", k->id);
	}
	else {
	  /* target knot */
	  printf_exit( !petrinet_set_start(p, k->id), 42, "Cant set start STELLE" );
	  do {
	    if (path_forward(p)) {
	      petrinet_path_out_flag(p,TRUE);
	    }
	  } while (branch_backward(p));
	  printf_log(1,"%i path between '%s' and '%s' found",
		       p->n_paths, p->start->id, p->target->id);
	}
      }
    }
    else {
      /* neither start nor target */
      assert(TRUE);
    }
  }

  result = xmlNewDoc(BAD_CAST "1.0");

  stylesheet = xmlNewDocNode(result, NULL, BAD_CAST "stylesheet", NULL);
  xmlSetProp(stylesheet, BAD_CAST "version", BAD_CAST "1.0");
  xmlDocSetRootElement(result, stylesheet);

  ns_xsl = xmlNewNs(stylesheet, BAD_CAST "http://www.w3.org/1999/XSL/Transform", BAD_CAST "xsl");
  xmlSetNs(stylesheet,ns_xsl);

  n_tmp = xmlNewChild(stylesheet, ns_xsl, BAD_CAST "output", NULL);
  xmlSetProp(n_tmp, BAD_CAST "method", BAD_CAST "xml");
  xmlSetProp(n_tmp, BAD_CAST "encoding", BAD_CAST "ISO-8859-1");

  template_xsl = xmlNewChild(stylesheet, ns_xsl, BAD_CAST "template", NULL);
  xmlSetProp(template_xsl, BAD_CAST "match", BAD_CAST "@*|node()");

  template_choose = xmlNewChild(template_xsl, ns_xsl, BAD_CAST "choose", NULL);
  /* add a dummy when, because empty */
  n_knot = xmlNewChild(template_choose, NULL, BAD_CAST "when", NULL);
  xmlSetProp(n_knot, BAD_CAST "test", BAD_CAST "false()");

  for (k=p->s; k->id; k++) {
    if (k->out == FALSE) {
      printf_log(3,"Filter STELLE '%s'",k->id);
      xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "@ID='%s'",k->id);
      n_knot = xmlNewChild(template_choose, NULL, BAD_CAST "when", NULL);
      xmlSetProp(n_knot, BAD_CAST "test", buffer);
    }
  }

  for (k=p->t; k->id; k++) {
    if (k->out == FALSE) {
      printf_log(3,"Filter TRANSITION '%s'",k->id);
      xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "@ID='%s'",k->id);
      n_knot = xmlNewChild(template_choose, NULL, BAD_CAST "when", NULL);
      xmlSetProp(n_knot, BAD_CAST "test", buffer);
    }
  }

  for (i=0; i < p->n_e; i++) {
    assert(p->edge[i].from);
    assert(p->edge[i].to);
    if  (p->edge[i].out == FALSE || p->edge[i].from->out == FALSE || p->edge[i].to->out == FALSE) {
      printf_log(3,"Filter relation '%s' -> '%s'", p->edge[i].from->id, p->edge[i].to->id);
      xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "@FROM='%s' and @TO='%s'",p->edge[i].from->id, p->edge[i].to->id);
      n_knot = xmlNewChild(template_choose, NULL, BAD_CAST "when", NULL);
      xmlSetProp(n_knot, BAD_CAST "test", buffer);
    }
  }

  n_knot = xmlNewChild(template_choose, NULL, BAD_CAST "otherwise", NULL);

  n_knot = xmlNewChild(n_knot, ns_xsl, BAD_CAST "copy", NULL);

  n_knot = xmlNewChild(n_knot, ns_xsl, BAD_CAST "apply-templates", NULL);
  xmlSetProp(n_knot,BAD_CAST "select", BAD_CAST "@*|node()");

  return result;
}
/* End of xml_pathnet() */


/*! compiles a DOM of pathtable

\param p pointer to used petrinet

\result complete DOM of pathtable
 */
xmlDocPtr
xml_pathtable(petrinet_t *p)
{
  /* 
  */
  int i = 0;
  xmlNodePtr n_tmp;
  xmlNodePtr pathtable;
  xmlNodePtr stylesheet;
  xmlNodePtr template_root;
  xmlNodePtr template_xsl;
  xmlNodePtr attribute_count;
  xmlNodePtr n_knot;
  xmlDocPtr result = NULL;
  xmlChar buffer[BUFFER_LENGTH];

  xmlNsPtr ns_xsl;

  if (p->start && p->target) {

    result = xmlNewDoc(BAD_CAST "1.0");

    stylesheet = xmlNewDocNode(result, NULL, BAD_CAST "stylesheet", NULL);
    xmlSetProp(stylesheet, BAD_CAST "version", BAD_CAST "1.0");
    xmlDocSetRootElement(result, stylesheet);

    ns_xsl = xmlNewNs(stylesheet, BAD_CAST "http://www.w3.org/1999/XSL/Transform", BAD_CAST "xsl");
    xmlSetNs(stylesheet,ns_xsl);

    n_tmp = xmlNewChild(stylesheet, ns_xsl, BAD_CAST "output", NULL);
    xmlSetProp(n_tmp, BAD_CAST "method", BAD_CAST "xml");
    xmlSetProp(n_tmp, BAD_CAST "encoding", BAD_CAST "ISO-8859-1");

    template_root = xmlNewChild(stylesheet, ns_xsl, BAD_CAST "template", NULL);
    xmlSetProp(template_root, BAD_CAST "match", BAD_CAST "/");

    pathtable = xmlNewChild(template_root, NULL, BAD_CAST "element", NULL);
    xmlSetProp(pathtable, BAD_CAST "name", BAD_CAST "PATHTABLE");

    attribute_count = xmlNewChild(pathtable, ns_xsl, BAD_CAST "attribute", NULL);
    xmlSetProp(attribute_count, BAD_CAST "name", BAD_CAST "COUNT");

    petrinet_all_out_flag(p,FALSE);

    do {
      if (path_forward(p)) {
/* 	petrinet_path_out_flag(p,TRUE); */
	i++;
	petrinet_path_out_flag(p,TRUE);
	if (i < 30) {		/* rescue counter */
	  if (!xml_path(p,pathtable,ns_xsl)) {
	    printf_log(1,"kk() failed");
	    return NULL;
	  }
	}
      }
    } while (branch_backward(p));

    xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "%i",i);
/*     xmlSetProp(pathtable, BAD_CAST "COUNT", buffer); */
    xmlNewTextChild(attribute_count, ns_xsl, BAD_CAST "text", buffer);

    printf_log(1,"%i path between '%s' and '%s' found",
		 p->n_paths, p->start->id, p->target->id);

    /* create rest of templates
     */
    template_xsl = xmlNewChild(stylesheet, ns_xsl, BAD_CAST "template", NULL);
    xmlSetProp(template_xsl, BAD_CAST "match", BAD_CAST "STELLE");

    n_knot = xmlNewChild(template_xsl, NULL, BAD_CAST "element", NULL);
    xmlSetProp(n_knot, BAD_CAST "name", BAD_CAST "STELLE");

    n_tmp = xmlNewChild(n_knot, ns_xsl, BAD_CAST "attribute", NULL);
    xmlSetProp(n_tmp, BAD_CAST "name", BAD_CAST "ID");
    n_tmp = xmlNewChild(n_tmp, ns_xsl, BAD_CAST "value-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "@ID");

    n_tmp = xmlNewChild(n_knot, ns_xsl, BAD_CAST "copy-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "NAME");
    n_tmp = xmlNewChild(n_knot, ns_xsl, BAD_CAST "copy-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "DESCRIPTION");

    template_xsl = xmlNewChild(stylesheet, ns_xsl, BAD_CAST "template", NULL);
    xmlSetProp(template_xsl, BAD_CAST "match", BAD_CAST "TRANSITION");

    n_knot = xmlNewChild(template_xsl, NULL, BAD_CAST "element", NULL);
    xmlSetProp(n_knot, BAD_CAST "name", BAD_CAST "TRANSITION");

    n_tmp = xmlNewChild(n_knot, ns_xsl, BAD_CAST "attribute", NULL);
    xmlSetProp(n_tmp, BAD_CAST "name", BAD_CAST "ID");
    n_tmp = xmlNewChild(n_tmp, ns_xsl, BAD_CAST "value-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "@ID");

    n_tmp = xmlNewChild(n_knot, ns_xsl, BAD_CAST "copy-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "NAME");
    n_tmp = xmlNewChild(n_knot, ns_xsl, BAD_CAST "copy-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "DESCRIPTION");

    n_tmp = xmlNewChild(n_knot, NULL, BAD_CAST "element", NULL);
    xmlSetProp(n_tmp, BAD_CAST "name", BAD_CAST "PARENT");
    n_tmp = xmlNewChild(n_tmp, ns_xsl, BAD_CAST "value-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "../NAME");

    template_xsl = xmlNewChild(stylesheet, ns_xsl, BAD_CAST "template", NULL);
    xmlSetProp(template_xsl, BAD_CAST "match", BAD_CAST "RELATION");

    n_knot = xmlNewChild(template_xsl, NULL, BAD_CAST "element", NULL);
    xmlSetProp(n_knot, BAD_CAST "name", BAD_CAST "RELATION");

    n_tmp = xmlNewChild(n_knot, ns_xsl, BAD_CAST "copy-of", NULL);
    xmlSetProp(n_tmp, BAD_CAST "select", BAD_CAST "DESCRIPTION");
  }

  return result;
}
/* End of xml_pathtable() */


/*! compiles a DOM of a path

\param p pointer to used petrinet
\param parent parent in DOM
\param ns_xsl namespace in DOM

\result complete DOM of a path
 */
xmlNodePtr
xml_path(petrinet_t *p, xmlNodePtr parent, xmlNsPtr ns_xsl)
{
  int i, m;
  knot_t *k;
  edge_t *e;
  xmlNodePtr result;
  xmlNodePtr step;
  xmlChar buffer[BUFFER_LENGTH];

/*   result = xmlNewDocNode(parent, NULL, "PATH", NULL); */
  result = xmlNewChild(parent, NULL, BAD_CAST "element", NULL);
  xmlSetProp(result, BAD_CAST "name", BAD_CAST "PATH");

  for (i=0, m=0; i <= p->l_path; i++) {
    k = p->path[i];

    /* insert DESCRIPTION of this RELATION */
    if (i>0) {
      e = edge_ptr4ptr(p,p->path[i-1],k);
      if (e) {
	step = xmlNewChild(result, ns_xsl, BAD_CAST "apply-templates",NULL);
	xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "//*[@FROM = '%s' and @TO = '%s']", e->from->id, e->to->id);
	xmlSetProp(step,BAD_CAST "select", buffer);
      }
    }
    step = xmlNewChild(result, ns_xsl, BAD_CAST "apply-templates",NULL);
    xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "//*[@ID = '%s']", k->id);
    xmlSetProp(step,BAD_CAST "select", buffer);
  }

  return result;
}
/* End of xml_path() */


