/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <limits.h>

/* 
 */
#include <libxml/HTMLtree.h>
#include <libxml/parser.h>

/* 
 */
#include <libxslt/xslt.h>
#include <libxslt/xsltutils.h>
#include <libxslt/transform.h>
#include <libxslt/variables.h>

#include <iconv.h>
#include <pcre.h>

#ifdef _MSC_VER
#include "config-win32.h"
#else
#include "../config.h"
#endif

/* 
 */
#include "basics.h"
#include "iconv_x.h"
#include "dom_x.h"
#include "dir_x.h"
#include "plain_x.h"
#include "make_x.h"
#include "calendar_x.h"
#include "tools.h"

#include "cxp_dtd.h"

#define RE_URL "(file|http|https|ftp|ftps|mailto):(//+[a-z0-9\\.\\-\\&\\#\\;\\:\\,\\+\\_%%\\~\\?\\!=\\@]+|[a-z0-9\\.\\-]+@)([/a-z0-9\\.\\-\\&\\#\\;\\,\\+\\_%%\\~\\?\\!=\\@])*"
#define RE_FIG "^[ \\t]*(Fig|Abb)\\.[ \\t]*([^ \\t:]+)[ \\t]*:*[ \\t]*(.+)*$"
#define RE_TAG "([A-Z]{3,}|[A-Za-zäÄöÖüÜß]{6,})"

static pcre *re_url  = NULL;
static pcre *re_fig  = NULL;
static pcre *re_tag  = NULL;

static xmlNodePtr
recognize_figures(xmlNodePtr node_cur);

static xmlNodePtr
recognize_urls(xmlNodePtr node_cur);

static xmlNodePtr
recognize_tags(xmlNodePtr node_tags, xmlNodePtr node_cur);

static xmlNodePtr
process_tags(xmlDocPtr doc_pie, xmlChar *attr_tags);

typedef enum {
  TYPE_HEADER,
  TYPE_PAR,
  TYPE_PRE,
  TYPE_LIST_ITEM,
  TYPE_LIST_ITEM_ENUM
} type_t;


xmlNsPtr ns_pie = NULL;

const xmlChar *ptr_pie = BAD_CAST "http://www.tenbusch.info/cxproc/pie";


/*! remove the line ends similar to perl's chomp()
*/
int
chomp(unsigned char *c)
{
  unsigned char *a;
  unsigned char *b;
  
  for ( a=b=c; *b==(unsigned char)' '; b++) { }
  
  for ( ; *b != (unsigned char)'\0'; b++) {
    switch (*b) {
    case '\n':
    case '\r':
      break;
    default:
      *a = *b;
      a++;
    }
  }
  *a = (unsigned char)'\0';
  for ( a--; a>c && *a==(unsigned char)' '; a--) { 
    *a = (unsigned char)'\0';
  }  

  return (a-c);
}
/* end of chomp() */


/*! Append the plain text from filehandle to the given node_parent

\param node_parent 
\param node_file node of file element with name and options
\param in input file handle

if in is NULL, the single text child node of node_file is used as single input paragraph

 */
BOOL_T
append_input(xmlNodePtr node_parent, xmlNodePtr node_file, FILE *in)
{
  BOOL_T header_root_defined = FALSE;
  int mode = 0;
  int mode_type;
  int mode_format;
  xmlChar *attr_type;
  xmlChar *attr_format;
  xmlChar *ptr_in;
  xmlNodePtr header[10];
  xmlNodePtr node_table = NULL;
  pcre *re_sep  = NULL;
  int erroffset;
  const char *error;

  header[0] = node_parent;
  header[1] = NULL;

  /*! detect additional settings attributes at node_file */

  mode_type = 2;		/* default */
  if ((attr_type = attribute_str(node_file,BAD_CAST "type"))) {
    if (xmlStrcasecmp(attr_type,BAD_CAST "line")==0) {
      mode_type = 1;
      printf_log(2,"Read line oriented");
    }
    else if (xmlStrcasecmp(attr_type,BAD_CAST "csv")==0) {
      xmlChar *attr_sep;

      mode_type = 3;
      printf_log(2,"Read CSV oriented");

      if ((attr_sep = attribute_str(node_file,BAD_CAST "sep"))) {
	printf_log(2,"Use Separator regexp '%s'", attr_sep);
	re_sep = pcre_compile((const char*)attr_sep,0,&error,&erroffset,NULL);
	/*!\todo error handling */
      }
      else {
	/*!\todo handling of quoted separators */
	attr_sep = BAD_CAST"[ \"\']*[\\t\\,\\;][ \"\']*";
	printf_log(2,"Use default Separator regexp '%s'", attr_sep);
	re_sep = pcre_compile((const char*)attr_sep,0,&error,&erroffset,NULL);
	/*!\todo error handling */
      }
    }
    else if (xmlStrcasecmp(attr_type,BAD_CAST "tabbed")==0) {
      mode_type = 4;
      printf_log(2,"Read Tabulator oriented");
      /*!\todo implement tabbed reading */
    }
  }

  mode_format = 0;
  if ((attr_format = attribute_str(node_file,BAD_CAST "format"))) {
    if (xmlStrcasecmp(attr_format,BAD_CAST "log")==0) {
      mode_format = 1;
      printf_log(2,"Read logfile format");
    }
  }

  /*
    substitutions with accurate UTF-8/XML strings
  */
#define STR_REPLACE(P,A,B)				\
  ptr_tmp = str_replace(P, BAD_CAST A, BAD_CAST B);	\
  if (ptr_tmp) {					\
    xmlFree(P);						\
    P = ptr_tmp;					\
  }

  /*! loop for reading paragraphs from file or buffer */

  while ((in==NULL 
	  && node_file->children->content!=NULL
	  && (ptr_in = xmlStrdup(node_file->children->content))!=NULL)
	 || 
	 (in!=NULL
	  && (ptr_in = fgetparUTF8(&mode,&mode_type,BUFFER_LENGTH-1,in))!=NULL)) {
    int depth_markup = 0;
    int level_hidden; 	/* counter for hidden level of a par */
    type_t type_node;
    xmlChar buffer_hidden[BUFFER_LENGTH];
    xmlChar *ptr_a, *ptr_b, *ptr_tmp;
    xmlNodePtr node_h;
    xmlNodePtr node_last = NULL;
    xmlNodePtr node_p;
    xmlNodePtr node_parent_4_append = node_parent;

    /*!\todo use language specific quotes according attribute in node_file */

    if (mode & 1) {
      /* this is a preformated paragraph */
      type_node = TYPE_PRE;
      STR_REPLACE(ptr_in,"&", "&amp;"); /* must be the first !!! */
      STR_REPLACE(ptr_in,"\"", "&quot;");
      STR_REPLACE(ptr_in,"\'", "&apos;");

      STR_REPLACE(ptr_in,"<", "&lt;");
      STR_REPLACE(ptr_in,">", "&gt;");

      node_last = node_parent_4_append = node_parent;
      ptr_a = NULL;
    }
    else {    
      level_hidden=0;
      for (ptr_b=ptr_in; *ptr_b == ';'; ptr_b++, level_hidden++) { };

      for ( ; *ptr_b == ' '; ptr_b++) { };

      if (mode_type==3) {
	/* in CSV mode empty lines are handled later */
      }
      else if (*ptr_b == '\0') {
	xmlFree(ptr_in);
	break;
      }
      ptr_a=ptr_b;
      switch (*ptr_b) {
      case '%':
	/*! set hidden level */
	level_hidden = 1;
	/*@fallthrough@*/
      case '*':
	type_node = TYPE_HEADER;
	for ( ; *ptr_a=='*' || *ptr_a=='%'; ptr_a++) { }
	break;
      case '+':
	type_node = TYPE_LIST_ITEM_ENUM;
	for ( ; *ptr_a=='+' || *ptr_a=='-'; ptr_a++) { }
	break;
      case '-':
	type_node = TYPE_LIST_ITEM;
	for ( ; *ptr_a=='+' || *ptr_a=='-'; ptr_a++) { }
	break; 
      default:
	type_node = TYPE_PAR;
      }
      depth_markup = ptr_a - ptr_b;
      for ( ; *ptr_a == ' '; ptr_a++) { };

      if (xmlStrlen(ptr_a) > 0) {
	ptr_a = xmlStrdup(ptr_a);
      }
      else {
	/* there is no content at ptr_a */
	ptr_a = NULL;
      }
      xmlFree(ptr_in);
      ptr_in = NULL;

      if (level_hidden) {
	xmlStrPrintf(buffer_hidden,BUFFER_LENGTH, BAD_CAST "%i",level_hidden);
      }

      if (mode_type==3) {
	/* CSV */
	xmlNodePtr node_row;
	xmlChar *ptr_value;

	if (ptr_a == NULL) {
	  node_table = NULL;
	  xmlFree(ptr_in);
	  continue;		/* empty lines are table separators */
	}
	if (node_table==NULL) {
	  node_table = xmlNewChild(node_parent, NULL, BAD_CAST "table", NULL);
	}
	node_parent_4_append = node_table;
	node_row = xmlNewChild(node_table, NULL, BAD_CAST "tr", NULL);

	if (re_sep) {
	  int rc;
	  int ovector[30];
	  int i;
	  xmlChar *ptr_substr;
          
	  /*\todo first replace leading and trailing text chars (" and ') */

	  for (ptr_substr=ptr_a, i=0; ptr_substr; i++) {

	    rc = pcre_exec(
			   re_sep,        /* result of pcre_compile() */
			   NULL,           /* we didn't study the pattern */
			   (const char*)ptr_substr,  /* the subject string */
			   strlen((const char*)ptr_substr),             /* the length of the subject string */
			   0,              /* start at offset 0 in the subject */
			   0,              /* default options */
			   ovector,        /* vector of integers for substring information */
			   30);            /* number of elements (NOT size in bytes) */

	    if (rc > -1) {
	      ptr_value = xmlStrndup(ptr_substr,ovector[0]);
	      STR_REPLACE(ptr_value,"&", "&amp;"); /* must be the first !!! */
	      STR_REPLACE(ptr_value,"\"", "&quot;");
	      STR_REPLACE(ptr_value,"\'", "&apos;");
	      xmlNewChild(node_row, NULL, BAD_CAST ((type_node == TYPE_HEADER) ? "th" : "td"), ptr_value); 
	      xmlFree(ptr_value);
	      ptr_substr += ovector[1];
	    }
	    else {
	      ptr_value = xmlStrdup(ptr_substr);
	      STR_REPLACE(ptr_value,"&", "&amp;"); /* must be the first !!! */
	      STR_REPLACE(ptr_value,"\"", "&quot;");
	      STR_REPLACE(ptr_value,"\'", "&apos;");
	      xmlNewChild(node_row, NULL, BAD_CAST ((type_node == TYPE_HEADER) ? "th" : "td"), ptr_value); 
	      xmlFree(ptr_value);
	      ptr_substr = NULL;
	    }
	  }
	}
	else {
	  /* no valid regexp prepared */
	  ptr_value = xmlStrdup(ptr_a);
	  STR_REPLACE(ptr_value,"&", "&amp;"); /* must be the first !!! */
	  STR_REPLACE(ptr_value,"\"", "&quot;");
	  STR_REPLACE(ptr_value,"\'", "&apos;");
	  xmlNewChild(node_row, NULL, BAD_CAST "td", ptr_value); 
	  xmlFree(ptr_value);
	}

	if (level_hidden) {
	  /* set the hidden attribute */
	  xmlSetProp(node_row, BAD_CAST "hidden", buffer_hidden);
	}
      }
      else {
	/* 
	 */
	STR_REPLACE(ptr_a,"&", "&amp;"); /* must be the first !!! */
	STR_REPLACE(ptr_a,"\"", "&quot;");
	STR_REPLACE(ptr_a,"\'", "&apos;");

	/* s. http://www.unicode.org */
	STR_REPLACE(ptr_a,"<<","&#x201C;");
	STR_REPLACE(ptr_a,">>","&#x201E;");
	STR_REPLACE(ptr_a,"<->","&#x2194;");
	STR_REPLACE(ptr_a,"->","&#x2192;");
	STR_REPLACE(ptr_a,"<-","&#x2190;");
	STR_REPLACE(ptr_a,"<", "&#x201B;");
	STR_REPLACE(ptr_a,">", "&#x201A;");

	node_last = node_parent_4_append = node_parent;
 
	/*!\bug dont modify node_parent_4_append for the first input element */
      }
    }

    if (mode_type==3) {
      /* CSV */
    }
    else if (type_node == TYPE_HEADER) {
      if (!header_root_defined) {
	/* the root header element of this input is not detected yet */
#if 1
	int i_d;
	for (i_d=0, node_last=node_parent; node_last; node_last = node_last->last) {

	  if (node_last->type == XML_ELEMENT_NODE) {
	    if (!xmlStrcasecmp(node_last->name, BAD_CAST "pie")) {
	    }
	    else if (!xmlStrcasecmp(node_last->name, BAD_CAST "section")) {
	      i_d++;
	      node_parent_4_append = node_last;
	      if (i_d == depth_markup - 1) {
		break;
	      }
	    }
	    else {
	      /* last node */
	      node_parent_4_append = node_last->parent;
	      break;
	    }
	  }
	}
	for ( ; i_d<depth_markup-1; i_d++) {
	  /* create empty SECTION elements */
	  node_parent_4_append = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "section", NULL);
	}

	if (node_parent_4_append) {
	}
	else {
	  int i;
	  node_parent_4_append = node_parent;
	  for (i=1; i<depth_markup; i++) {
	    /* create empty SECTION elements */
	    node_parent_4_append = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "section", NULL);
	  }
	}
#else
	int i;
	for (i=1; i<d; i++) {
	  /* create empty SECTION elements */
	  node_parent_4_append = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "section", NULL);
	}
#endif
	header_root_defined = TRUE;
      }
      for ( ; node_last; node_last = node_last->last) {
	if (node_last->type == XML_ELEMENT_NODE
	    && xmlStrEqual(node_last->name, BAD_CAST "section")) {
	  if (depth_markup<2)
	    break; /* if node_last has the same number of '*' -> sibling */
	  node_parent_4_append = node_last;
	  depth_markup--;
	}
      }
      node_p = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "section", NULL);
      node_h = xmlNewChild(node_p, NULL, BAD_CAST "h", ptr_a);
      if (level_hidden) {
	xmlSetProp(node_h, BAD_CAST "hidden", buffer_hidden);
      }
    }
    else if (type_node == TYPE_LIST_ITEM || type_node == TYPE_LIST_ITEM_ENUM) {
      /* search for the last SECTION */
      if (header_root_defined) {
	for ( ; node_last; node_last = node_last->last) {
	  if (node_last->type == XML_ELEMENT_NODE) {
	    if (xmlStrEqual(node_last->name, BAD_CAST "section")) {
	      node_parent_4_append = node_last;
	    }
	  }
	}
      }
      /* search for the LIST from the last SECTION */
      for ( node_last = node_parent_4_append; node_last; node_last = node_last->last) {
	if (node_last->type == XML_ELEMENT_NODE
	    && xmlStrEqual(node_last->name, BAD_CAST "list")) {
	  if (depth_markup<1)
	    break; /* if node_last has the same number of '-' -> sibling */
	  node_parent_4_append = node_last;
	  depth_markup--;
	}
      }
      for ( ; depth_markup>0; depth_markup--) {
	/* create the needed LISTs */
	node_parent_4_append = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "list", NULL);
	if (type_node == TYPE_LIST_ITEM_ENUM)
	  xmlSetProp(node_parent_4_append, BAD_CAST "enum", BAD_CAST "yes");
      }

      /* detect if node_parent_4_append is a LIST */
      if (node_parent_4_append->type == XML_ELEMENT_NODE
	  && xmlStrEqual(node_parent_4_append->name, BAD_CAST "list")) {
      }
      else {
	node_parent_4_append = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "list", NULL);
	if (type_node == TYPE_LIST_ITEM_ENUM)
	  xmlSetProp(node_parent_4_append, BAD_CAST "ENUM", BAD_CAST "yes");
      }
      node_p = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "p", ptr_a); 
      if (level_hidden) {
	/* set the hidden attribute */
	xmlSetProp(node_p, BAD_CAST "hidden", buffer_hidden);
      }
    }
    else {
      if (header_root_defined) {
	/* search for the last SECTION */
	for ( ; node_last; node_last = node_last->last) {
	  if (IS_NODE(node_last,BAD_CAST"section")) {
	    node_parent_4_append = node_last;
	  }
	}
      }
      if (type_node == TYPE_PRE) {
	xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "pre", ptr_in);
      }
      else {
	if (mode_format==0) {
	  /* plain text format */
	  node_p = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "p", ptr_a); 
	}
	else {
	  /* logfile format */
	  xmlChar *ptr_sep, *ptr_b, *ptr_c;
	  xmlNodePtr node_link;

	  for ( ptr_sep = ptr_a; isspace(*ptr_sep); ptr_sep++) {}
	  if (isdigit(ptr_sep[0]) && isdigit(ptr_sep[1]) && isdigit(ptr_sep[2]) && isdigit(ptr_sep[3])) {
	    /* 4 leading year digits */
	    for ( ptr_b = ptr_sep; iscal(*ptr_sep) || *ptr_sep==',' || *ptr_sep=='#'; ptr_sep++) {}
	    for ( ptr_c = ptr_sep; isspace(*ptr_sep); ptr_sep++) {}

	    if (ptr_c-ptr_b > 3) {
	      xmlChar *str_date = xmlStrndup(ptr_b,ptr_c-ptr_b);
	      int int_free;

	      /* detect holiday markup */
	      for ( int_free=0; *ptr_sep=='+'; ptr_sep++) {
		int_free++;
	      }
	      for ( ; isspace(*ptr_sep); ptr_sep++) {}

	      node_p = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "p", NULL); 
	      node_link = xmlNewChild(node_p, NULL, BAD_CAST "link", ptr_sep); 
	      xmlSetProp(node_link, BAD_CAST "date", str_date);
	      xmlFree(str_date);
	      if (int_free > 1) {
		xmlSetProp(node_link, BAD_CAST "free", BAD_CAST "yes");
	      }
	      else if (int_free > 0) {
		xmlSetProp(node_link, BAD_CAST "holiday", BAD_CAST "yes");
	      }
	    }
	    else {
	      node_p = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "p", ptr_a); 
	    }
	  }
	  else {
	    node_p = xmlNewChild(node_parent_4_append, ns_pie, BAD_CAST "p", ptr_a); 
	  }
	}

	if (level_hidden) {
	  /* set the hidden attribute */
	  xmlSetProp(node_p, BAD_CAST "hidden", buffer_hidden);
	}
      }
    }
    if (ptr_a) xmlFree(ptr_a);
    xmlFree(ptr_in);
    if (in==NULL) {
      /* input is the single text child node of node_file */
      break;
    }
  }
  /*!\todo STR_REPLACE(ptr_in,"--", "&#x2013;"); */
#undef STR_REPLACE

  if (re_sep) {
    pcre_free(re_sep);
    re_sep = NULL;
  }
  return TRUE;
}
/* end of append_input() */


/*! process the PIE child instructions of node_make_pie
 */
xmlDocPtr
process_pie(xmlNodePtr node_make_pie)
{
  xmlNodePtr node_file;
  xmlNodePtr node_meta;
  xmlDocPtr result;
  xmlNodePtr node_pie;
  xmlChar *ptr_attr;
  xmlChar buffer[BUFFER_LENGTH];

  printf_log(1,"Processing PIE files");

  result = xmlNewDoc(BAD_CAST "1.0");
  result->encoding = xmlStrdup(BAD_CAST "UTF-8");
  node_pie = xmlNewDocNode(result, NULL, NAME_PIE, NULL); 
  if (node_make_pie->children == NULL) {
    return result;
  }

  /*!\todo problems in XSL when namespace without prefix is set */
#if 0
  if (ns_pie) {
    xmlSetNs(node_pie,ns_pie);
  }
  else {
    ns_pie = xmlNewNs(node_pie, ptr_pie, NULL); /* no prefix string for this ns */
  }
#endif

  xmlSetProp(node_pie, BAD_CAST "class", BAD_CAST "article");
  xmlDocSetRootElement(result, node_pie);

  node_meta = xmlNewChild(node_pie, NULL, BAD_CAST "meta", NULL);
  xmlAddChild(node_meta,xmlCopyNode(node_make_pie,1));
  /* Get the current time. */
  xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "%i", (int)time(NULL));
  xmlSetProp(node_meta, BAD_CAST "mtime", BAD_CAST buffer);

  for (node_file=node_make_pie->children; 
       node_file;
       node_file = node_file->next) {
    if (node_file->type == XML_ELEMENT_NODE) {
      xmlChar *name_file = file_normalize(attribute_str(node_file,BAD_CAST "name"),NULL);
      if (IS_NODE_XML(node_file)) {
	xmlDocPtr doc_pie;
	doc_pie = process_xml(node_file);	  /*  call */
	if (doc_pie) {
	  xmlNodePtr node_r = xmlDocGetRootElement(doc_pie);
	  if (node_r !=NULL && node_r->type == XML_ELEMENT_NODE) {
	    printf_log(1,"Appending PIE");
	    /* FILE is a PIE xml */
	    /* its really a PIE, copy all childs */
	    for (node_r = node_r->children; node_r; node_r = node_r->next) {
	      xmlNodePtr node_dummy;
	      xmlNodePtr node_copy;
	      node_dummy = xmlNewChild(node_pie, NULL, BAD_CAST "dummy", NULL);
	      node_copy = xmlCopyNode(node_r,1);
	      xmlReplaceNode(node_dummy,node_copy); 
	      xmlFreeNode(node_dummy);
	    }
	  }
	  xmlFreeDoc(doc_pie);
	}
      }
      else if (IS_NODE_PLAIN(node_file)) {

	if (name_file==NULL) {
	  /* there is no named file */
	  printf_log(2,"Use text child node");
 	  append_input(node_pie,node_file,NULL);
	}
	else if (xmlStrEqual(name_file, (const xmlChar *) "-")) {
	  /* name for stdin */
	  append_input(node_pie,node_file,stdin);
	}
	else {
	  FILE *in;
	  printf_log(1,"Read pie file '%s'", name_file);
	  if ((in = fopen((char *)name_file,"r"))) {
	    /* FILE is a txt2x file */
	    append_input(node_pie,node_file,in);
	    fclose(in);
	  }
	  else {
	    printf_log(1,"Cant open pie file '%s'", name_file);
	  }
	}
      }
      if (name_file) xmlFree(name_file);
    }
  }
  ptr_attr = attribute_str(node_make_pie,BAD_CAST "figure");
  if (ptr_attr !=NULL && xmlStrcasecmp(ptr_attr,BAD_CAST "no")==0) {
    /* explicit "no" */
  }
  else {
    /* global figure recognition */
    printf_log(1,"Recognize Figures");
    recognize_figures(node_pie);	
  }

  ptr_attr = attribute_str(node_make_pie,BAD_CAST "url");
  if (ptr_attr !=NULL && xmlStrcasecmp(ptr_attr,BAD_CAST "no")==0) {
    /* explicit "no" */
  }
  else {
    /* global link recognition */
    printf_log(1,"Recognize URLs");
    /*!\todo use an attribute for regexp */
    recognize_urls(node_pie);	
  }

  ptr_attr = attribute_str(node_make_pie,BAD_CAST "tags");
  if (ptr_attr==NULL || xmlStrcasecmp(ptr_attr,BAD_CAST "no")==0) {
    /* implicit or explicit "no" */
  }
  else {
    /* global tag recognition */
    printf_log(1,"Recognize Tags");
    /*!\todo use an attribute for regexp */
    process_tags(result,ptr_attr);
  }

  /*!\todo global cite recognition in scientific text */

  return result;
}


/*! process a PLAIN instruction
*/
xmlChar *
process_plain(xmlNodePtr n)
{
  xmlChar *result_str = NULL;

  if (n->children) {
    if (n->children == n->last && n->children->type == XML_TEXT_NODE) {
      /*
	only one TEXT_NODE children, return pointer to a text copy
      */
      result_str = xmlStrdup(n->children->content);
    }
    else {
      /*
	there are additional or other types of childrens
      */
      xmlNodePtr c;
      if ((c = child_element(n,NAME_XML))) {
	/* there is a XML instruction, only first XML!! */
	xmlDocPtr doc_xml = process_xml(c);
	if (doc_xml) {
	  result_str = BAD_CAST process_xsl(doc_xml,n);
	  xmlFreeDoc(doc_xml); 
	}
      }
      else if ((c = child_element(n,NAME_IMAGE))) {
	/* there is a number of child elements of IMAGE */

	xmlChar *p;
	FILE *file_comment;
	long int size_file;
	xmlChar *name_file = attribute_str(c,BAD_CAST "src");

	/* derive name of comment file from image file */
	for ( p = name_file + xmlStrlen(name_file); p > name_file && *p != '.'; p-- ) {
	  ;
	}
	*(p+1) = '\0';

	result_str = xmlStrcat(name_file,BAD_CAST "txt");
	if (result_str != NULL
	    && type_is_readableUTF8(result_str,FALSE,&size_file)
	    && size_file > 0
	    && (file_comment = fopen((char *)result_str,"r"))!=NULL) {
	    unsigned char *b, *c, *d;
	    BOOL_T flag_skipped;
	    int input_length;

	    b = xmlMalloc(size_file+3);
	    b[0] = '\'';
	    input_length = fread((void *)(b+1),1,size_file,file_comment);
	    if (input_length != size_file) {
	      printf_log(1,"Read error '%s'", result_str);
	      return NULL;
	    }
	    printf_log(2,"Use %i byte from '%s'", size_file, result_str);
	    b[input_length+1] = '\'';
	    b[input_length+2] = '\0';

	    /*\todo remove line breaks etc. normalize_space() ? */
	    /*
	      remove line breaks etc.
	     */
	    for (flag_skipped=FALSE, d=b+1, c=b+1; *c; c++) {
 	      if (isspace(*c)) {
		if (flag_skipped) {
		  /* char before was also skipped,  */
		}
		else {
		  *d = ' ';
		  d++;
		  flag_skipped=TRUE;
		}
	      }
	      else {
		*d = *c;
		d++;
		flag_skipped=FALSE;
	      }
	    }
	    /*!\todo remove spaces at end of b */
/* 	    for ( ; *d==' ' && d>b; d--) { } */
	    *d = '\0';
	    printf_log(3,"Comment '%s'", b);
	    result_str = b;
	    fclose(file_comment);
	}
	else {
	  printf_log(2,"No valid comment file '%s' found", result_str);
	  result_str = xmlStrdup(BAD_CAST"''");
	}
      }
      else if (child_element(n,NAME_PLAIN)) {
	/* there is a number of child elements of PLAIN */
	for (c=n->children; c; c=c->next) {
	  if (IS_NODE_PLAIN(c)) {
	    xmlChar *result_plain = process_plain(c);
	    if (result_str) {
	      /* concatenation of PLAIN results */
	      xmlChar *result_str_old = result_str;
	      result_str = xmlStrncatNew(result_str,result_plain,xmlStrlen(result_plain));
	      xmlFree(result_str_old);
	      xmlFree(result_plain);
	    }
	    else {
	      result_str = result_plain;
	    }
	  }
	  else {
	    printf_log(1,"No valid element");
	  }
	}
      }
    }
    save_file(NULL,result_str,n);
  }
  else {
    /*
      element without children, read the attribute named file
    */
    long int size_file;
    xmlChar *name_file = file_normalize(attribute_str(n,BAD_CAST "name"),NULL);

    if (name_file
	&& type_is_readableUTF8(name_file,FALSE,&size_file)
	&& size_file > 0) {
      FILE *in;
      int n_f = (int)size_file;
      unsigned char *buffer_in  = (unsigned char *) xmlMemMalloc(n_f+1);
      /*!\bug  file_size() is not impemented for URL sources */
      printf_exit(buffer_in==NULL,22,"No more memory");

      if ((in = fopen((char *)name_file,"r"))) {
	printf_log(1,"Read %i bytes from plain file '%s'", n_f, name_file);
	if (fread(buffer_in,n_f,1,in) == 1) {
	  buffer_in[n_f] = '\0';
	  if (xmlCheckUTF8(buffer_in)) {
	    result_str = xmlStrdup(buffer_in);
	  }
	  else {
	    int l;
	    result_str = (xmlChar *) xmlMemMalloc(n_f*2); /*!\todo calculate length more accurate */
	    printf_exit(result_str==NULL,22,"No more memory");
	    l = isolat1ToUTF8(result_str,&n_f,buffer_in,&n_f);
	    if (l < 1) {
	      printf_log(1,"Problem with entry");
	    }
	    result_str[l] = '\0';
	  }
	}
	fclose(in);
      }
      else {
	printf_log(1,"Cant read plain file '%s'", name_file);
      }
      xmlMemFree(buffer_in);
    }
    if (name_file) xmlFree(name_file);
  }
  return result_str;
}
/* end of process_plain() */



/*! splits an UTF-8 string into a list of text and link element nodes
*/
xmlNodePtr
split_into_links(const xmlChar *ptr_origin)
{
  int int_strlen;
  xmlNodePtr node_result = NULL;

  if (ptr_origin != NULL && (int_strlen = xmlStrlen(ptr_origin)) > 0) {
    int rc;
    int ovector[30];
    xmlNodePtr node_link;  
    xmlNodePtr node_postfix;

    rc = pcre_exec(
		   re_url,        /* result of pcre_compile() */
		   NULL,           /* we didn't study the pattern */
		   (const char*)ptr_origin,  /* the subject string */
		   strlen((const char*)ptr_origin),             /* the length of the subject string */
		   0,              /* start at offset 0 in the subject */
		   0,              /* default options */
		   ovector,        /* vector of integers for substring information */
		   30);            /* number of elements (NOT size in bytes) */

    if (rc > -1 && ovector[1] - ovector[0] > 3) {
      /* 
	 the regexp match, assemble node list with a common dummy
	 element node
       */
      xmlChar *ptr_url;

      ptr_url = xmlStrndup(ptr_origin + ovector[0], ovector[1] - ovector[0]);
      printf_log(3,"URL '%s' (%i..%i) in '%s'", ptr_url,ovector[0],ovector[1],ptr_origin);

      node_result = xmlNewNode(NULL,BAD_CAST "dummy");

      if (ovector[0] > 0) {
	/* the content starts with text	*/
	xmlAddChild(node_result,xmlNewText(xmlStrndup(ptr_origin,ovector[0])));
      }

      node_link = xmlNewTextChild(node_result,NULL,BAD_CAST "link",ptr_url);
      xmlSetProp(node_link, BAD_CAST "href",ptr_url);

      if (int_strlen > ovector[1]) {
	/*
	  the content ends with text
	*/
	/* recursion */
	node_postfix = split_into_links(ptr_origin + ovector[1]);
	if (node_postfix) {
	  xmlAddChild(node_result,node_postfix->children);
	  xmlFree(node_postfix);
	}
	else {
	  xmlAddChild(node_result,xmlNewText(xmlStrdup(ptr_origin + ovector[1])));
	}
      }
      xmlFree(ptr_url);
    }
  }
  return node_result;
}
/* end of split_into_links() */


/*! derive a sequence of text and LINK nodes from node
 */
xmlNodePtr
recognize_urls(xmlNodePtr node_cur)
{
  if (re_url==NULL) {
    int erroffset;
    const char *error;
    printf_log(2,"Initialize URL regexp '%s'", RE_URL);
    re_url = pcre_compile((const char*)RE_URL,PCRE_CASELESS,&error,&erroffset,NULL);
    /*!\todo error handling */
  }

  if (IS_NODE(node_cur,NULL) && !IS_NODE(node_cur,BAD_CAST"pre") && (node_cur->ns==NULL || node_cur->ns==ns_pie)) {
    xmlNodePtr node_child;  
    xmlNodePtr node_for_append;  
    
    for (node_child = node_cur->children, node_for_append=node_child; node_child; node_child=node_child->next) {

      if (xmlNodeIsText(node_child)) {
	/* node_child is a text node */
	xmlNodePtr node_replace;

	node_replace = split_into_links(node_child->content);
	if (node_replace) {
	  if (IS_NODE(node_cur,BAD_CAST"link")) {
	    if (attribute_str(node_cur,BAD_CAST"href")==NULL) {
	      /* URL inside a link with no href attribute */
	      xmlNodePtr node_first_link = child_element(node_replace,BAD_CAST"link");
	      assert(node_first_link);
	      xmlSetProp(node_cur, BAD_CAST "href",attribute_str(node_first_link,BAD_CAST"href"));
	    }
	    xmlFreeNodeList(node_replace);
	  }
	  else {
	    /* there is a result list */
	    xmlNodePtr node_release = node_child;
	    xmlNodePtr node_i;

	    /* node_child = xmlReplaceNode(node_child,node_replace->children); */

	    xmlSetListDoc(node_replace->children,node_release->doc);

	    node_release->parent->children = node_replace->children;
	    node_release->parent->last     = node_replace->last;

	    for (node_i=node_replace->children; node_i; node_i=node_i->next) {
	      node_i->parent = node_release->parent;
	    }

	    xmlFreeNode(node_release);
	    xmlFree(node_replace->name);
	    xmlFree(node_replace);
	  }
	}
      }
      else {
	recognize_urls(node_child);
      }
    }      
  }
  return NULL;
}
/* End of recognize_urls() */


/*! derive a sequence of text and LINK nodes from node
 */
xmlNodePtr
recognize_figures(xmlNodePtr node_cur)
{
  if (re_fig==NULL) {
    int erroffset;
    const char *error;
    /* compile regexp first */
    printf_log(2,"Initialize Figure regexp '%s'", RE_FIG);
    re_fig = pcre_compile((const char*)RE_FIG,PCRE_CASELESS,&error,&erroffset,NULL);
    /*!\todo error handling */
  }

  if (IS_NODE(node_cur,NULL)) {
    xmlNodePtr node_child;  
    xmlNodePtr node_for_append;  
    int int_strlen;

    if (IS_NODE(node_cur,BAD_CAST"p")) {
      if ((node_child = node_cur->children) != NULL 
	&& xmlNodeIsText(node_child) 
	&& node_child->content != NULL
	&& (int_strlen = xmlStrlen(node_child->content)) > 0) {
	  int rc;
	  int ovector[30];
	  xmlChar *ptr_substr = node_child->content;

	  rc = pcre_exec(
	    re_fig,        /* result of pcre_compile() */
	    NULL,           /* we didn't study the pattern */
	    (const char*)ptr_substr,  /* the subject string */
	    strlen((const char*)ptr_substr),             /* the length of the subject string */
	    0,              /* start at offset 0 in the subject */
	    0,              /* default options */
	    ovector,        /* vector of integers for substring information */
	    30);            /* number of elements (NOT size in bytes) */

	  if (rc > 2 && ovector[1] - ovector[0] > 3) {
	    xmlChar *ptr_release;
	    xmlNodePtr node_release;

	    ptr_release = xmlStrndup(ptr_substr + ovector[4], ovector[5] - ovector[4]);
	    printf_log(3,"fig '%s' (%i..%i) in '%s'",ptr_release,ovector[0],ovector[1],ptr_substr);

	    xmlNodeSetName(node_cur, BAD_CAST "fig");
	    node_for_append = xmlNewChild(node_cur,NULL,BAD_CAST"img",NULL);
	    xmlSetProp(node_for_append, BAD_CAST "src",ptr_release);
	    xmlFree(ptr_release);

	    if (rc > 3 && ovector[7] - ovector[6] > 0) {
	      ptr_release = xmlStrndup(ptr_substr + ovector[6], ovector[7] - ovector[6]);
	      node_for_append = xmlNewChild(node_cur,NULL,BAD_CAST"h",ptr_release);
	      xmlFree(ptr_release);
	    }
	    node_release = node_cur->children;
	    xmlUnlinkNode(node_release);
	    xmlFreeNode(node_release);
	  }
	  /* else ignore empty p elements */
      }
    }
    else {
      for (node_child = node_cur->children; node_child; node_child=node_child->next) {
	recognize_figures(node_child);
      }
    }
  }
  return NULL;
}
/* End of recognize_figures() */

/*! process the 
 */
xmlNodePtr
process_tags(xmlDocPtr doc_pie, xmlChar *attr_tags)
{
  xmlNodePtr node_tags;
  xmlNodePtr node_word;
  int erroffset;
  const char *error;

  if (re_tag) pcre_free(re_tag);

  if (xmlStrEqual(attr_tags,BAD_CAST"yes")) {
    /* use default regexp */
    printf_log(2,"Initialize tag regexp with default '%s'", RE_TAG);
    re_tag = pcre_compile((const char*)RE_TAG,0,&error,&erroffset,NULL);
    /*!\todo error handling */
  }
  else {
    /* compile regexp first */
    printf_log(2,"Initialize tag regexp '%s'", attr_tags);
    re_tag = pcre_compile((const char*)attr_tags,0,&error,&erroffset,NULL);
    /*!\todo error handling */
  }

  node_tags = xmlNewNode(NULL,BAD_CAST"tags");

  xmlAddChild(child_element(doc_pie->children,NAME_META),node_tags);
  recognize_tags(node_tags,doc_pie->children);

  for (node_word = node_tags->children; node_word; node_word=node_word->next) {
    /* set counter attribute, fix misused pointer to properties !!!DANGER!!! */
    xmlChar buffer[100];
    if ((int)node_word->properties > 0) {
      xmlStrPrintf(buffer,sizeof(buffer),BAD_CAST"%i",1+(int)node_word->properties);
      node_word->properties = NULL;
      xmlSetProp(node_word,BAD_CAST"count",buffer);
    }
    else {
      node_word->properties = NULL;
    }
  }

  return node_tags;
}


/*! recognize all remarkable text tags and append them as childs to node_tags
 */
xmlNodePtr
recognize_tags(xmlNodePtr node_tags, xmlNodePtr node_cur)
{
  assert(re_tag);

  if (IS_NODE(node_cur,NULL)) {
    xmlNodePtr node_child;  
    xmlNodePtr node_for_append;  

    if (!IS_NODE_META(node_cur)) {
      for (node_child = node_cur->children, node_for_append=node_child; node_child; node_child=node_child->next) {
	if (xmlNodeIsText(node_child) && node_child->content) {
	  int rc;
	  int ovector[30];
	  int i;
	  xmlChar *ptr_substr;
          
	  for (ptr_substr=node_child->content, i=0; ptr_substr; i++) {
	    rc = pcre_exec(
			   re_tag,        /* result of pcre_compile() */
			   NULL,           /* we didn't study the pattern */
			   (const char*)ptr_substr,  /* the subject string */
			   strlen((const char*)ptr_substr),             /* the length of the subject string */
			   0,              /* start at offset 0 in the subject */
			   0,              /* default options */
			   ovector,        /* vector of integers for substring information */
			   30);            /* number of elements (NOT size in bytes) */
	    if (rc > -1) {
	      xmlChar *ptr_tag;
	      xmlNodePtr node_word;

	      printf_log(3,"tag %i..%i in '%s'", ovector[0],ovector[1],ptr_substr);
	      ptr_tag = xmlStrndup(ptr_substr + ovector[0], ovector[1] - ovector[0]);

	      /* search in tags for this value */
	      for (node_word = node_tags->children; ; node_word=node_word->next) {

		if (node_word != NULL
		    && node_word->children != NULL
		    && node_word->children->content != NULL
		    && xmlStrEqual(node_word->children->content,ptr_tag)) {
		  /* increment counter, misuse pointer to properties !!!DANGER!!! */
		  node_word->properties = 1 + (int)node_word->properties;
		  break;
		}

		if (node_word==NULL || node_word->next==NULL) {
		  /* add new tag */
		  node_for_append = xmlAddChild(node_tags,xmlNewNode(NULL, BAD_CAST "word"));
		  xmlNodeSetContent(node_for_append,ptr_tag);
		  node_for_append->properties = 0;
		  break;
		}
	      }	
	      xmlFree(ptr_tag);

	      if (*(ptr_substr + ovector[1]) == '\0') {
		/* there are no trailing chars */
		ptr_substr = NULL;
	      }
	      else {
		ptr_substr += ovector[1];
	      }
	    }
	    else {
	      ptr_substr = NULL;
	    }
	  }
	}
	else {
	  recognize_tags(node_tags, node_child);
	}
      }      
    }
  }
  return NULL;
}
/* End of recognize_tags() */



/*!\todo process_quotes() for standardized quotes */
